# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)
"""Tool helping to port an addon or missing commits of an addon from one branch
to another.

If the addon does not exist on the target branch, it will assist the user
in the migration, following the OCA migration guide.

If the addon already exists on the target branch, it will retrieve missing
commits to port. If a Pull Request exists for a missing commit, it will be
ported with all its commits if they were not yet (fully) ported.

To check if an addon could be migrated or to get eligible commits to port:

    $ export GITHUB_TOKEN=<token>
    $ oca-port origin/13.0 origin/14.0 shopfloor --verbose

To effectively migrate the addon or port its commits, use the `--destination` option:

    $ oca-port origin/13.0 origin/14.0 shopfloor --destination camptocamp/14-mig-shopfloor

Note that you can omit the remote to work on local branches if needed:

    $ oca-port 13.0 camptocamp/14-dev shopfloor

NOTE: if the source branch is a local one, cache will be readonly and
won't be updated as commits data coming from such branch cannot be trust.

The organization used to perform API requests to look for Pull Requests data
is the one defined through `--upstream-org` parameter (defaults to 'OCA').
So you can work on local source and target branches while performing API requests
on the relevant upstream organization:

    $ oca-port 14.0 16.0 shopfloor --upstream-org camptocamp

Migration of addon
------------------

The tool follows the usual OCA migration guide to port commits of an addon,
and will invite the user to fullfill the mentionned steps that can't be
performed automatically.

Port of commits/Pull Requests
-----------------------------

The tool will ask the user if he wants to open draft pull requests against
the upstream repository.

If there are several Pull Requests to port, it will ask the user if he wants to
base the next PR on the previous one, allowing the user to cumulate ported PRs
in one branch and creating a draft PR against the upstream repository with all
of them.
"""

import click

from ..app import App
from ..exceptions import ForkValueError, RemoteBranchValueError
from ..utils.misc import bcolors as bc


@click.command()
@click.argument("source", required=True)
@click.argument("target", required=True)
@click.argument("addon_path", required=True)
@click.option(
    "--destination",
    help=("Git reference where work will be pushed, e.g. 'camptocamp/16.0-dev'."),
)
@click.option(
    "--source-version",
    help="Source Odoo version. To set if it cannot be detected from 'source'.",
)
@click.option(
    "--target-version",
    help="Target Odoo version. To set if it cannot be detected from 'target'.",
)
@click.option("--repo-name", help="Repository name, e.g. 'server-tools'.")
@click.option(
    "--upstream-org",
    default="OCA",
    show_default=True,
    help="Upstream organization name. Used for API requests.",
)
@click.option("--verbose", is_flag=True, help="List the commits of Pull Requests.")
@click.option(
    "--non-interactive", is_flag=True, help="Disable all interactive prompts."
)
@click.option("--dry-run", is_flag=True, help="Print results, no nothing.")
@click.option(
    "--output",
    help=(
        "Returns the result in a given format. "
        "This implies the `--non-interactive` option automatically. "
        "Possibles values are: 'json'."
    ),
)
@click.option("--fetch", is_flag=True, help="Fetch remote branches from upstream.")
@click.option("--no-cache", is_flag=True, help="Disable user's cache.")
@click.option("--clear-cache", is_flag=True, help="Clear the user's cache.")
def main(
    addon_path: str,
    source: str,
    target: str,
    destination: str,
    source_version: str,
    target_version: str,
    repo_name: str,
    upstream_org: str,
    verbose: bool,
    non_interactive: bool,
    output: str,
    fetch: bool,
    no_cache: bool,
    clear_cache: bool,
    dry_run: bool,
):
    """Migrate ADDON from SOURCE to TARGET or list Pull Requests to port.

        E.g.:

        $ oca-port origin/14.0 origin/16.0 auditlog

        Migration:

            Assist the user in the migration of the addon, following the OCA guidelines.

        Port of Pull Requests (missing commits):

            The PRs are found from SOURCE commits that do not exist in TARGET.
    The user will be asked if he wants to port them.
    """
    try:
        app = App(
            addon_path=addon_path,
            source=source,
            target=target,
            destination=destination,
            source_version=source_version,
            target_version=target_version,
            repo_name=repo_name,
            upstream_org=upstream_org,
            verbose=verbose,
            non_interactive=non_interactive,
            output=output,
            fetch=fetch,
            no_cache=no_cache,
            clear_cache=clear_cache,
            dry_run=dry_run,
            cli=True,
        )
    except ForkValueError as exc:
        error_msg = prepare_remote_error_msg(*exc.args)
        error_msg += (
            "\n\nYou can change the GitHub organization with the "
            f"{bc.DIM}--user-org{bc.END} option."
        )
        raise click.ClickException(error_msg) from exc
    except RemoteBranchValueError as exc:
        error_msg = prepare_remote_error_msg(*exc.args)
        raise click.ClickException(error_msg) from exc
    except ValueError as exc:
        raise click.ClickException(exc) from exc
    # Run the app
    try:
        app.run()
    except ValueError as exc:
        raise click.ClickException(exc) from exc


def prepare_remote_error_msg(repo_name, remote):
    return (
        f"No remote {bc.FAIL}{remote}{bc.END} in the current repository.\n"
        "To add it:\n"
        "\t# This mode requires an SSH key in the GitHub account\n"
        f"\t{bc.DIM}$ git remote add {remote} "
        f"git@github.com:{remote}/{repo_name}.git{bc.END}\n"
        "   Or:\n"
        "\t# This will require to enter user/password each time\n"
        f"\t{bc.DIM}$ git remote add {remote} "
        f"https://github.com/{remote}/{repo_name}.git{bc.END}"
    )


if __name__ == "__main__":
    main()



# Copyright 2023 Camptocamp SA
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl)

from . import common

from oca_port.utils import cache


class TestUserCache(common.CommonCase):
    def setUp(self):
        super().setUp()
        app = self._create_app(self.source1, self.target1, upstream_org="TEST")
        self.cache = cache.UserCache(app)
        # As source branch as no organization, cache is by default readonly.
        # Unset this to run tests.
        self.cache.readonly = False
        # Clear the cache before each test
        self.cache.clear()

    def tearDown(self):
        # Clear the cache after each test
        self.cache.clear()

    def test_commit_ported(self):
        sha = "TEST"
        self.assertFalse(self.cache.is_commit_ported(sha))
        self.cache.mark_commit_as_ported(sha)
        self.assertTrue(self.cache.is_commit_ported(sha))

    def test_commit_pr(self):
        sha = "TEST"
        pr_data = {
            "number": 10,
            "title": "TEST",
        }
        self.assertFalse(self.cache.get_pr_from_commit(sha))
        self.cache.store_commit_pr(sha, pr_data)
        self.assertDictEqual(self.cache.get_pr_from_commit(sha), pr_data)

    def test_commit_files(self):
        sha = "TEST"
        files = ["a/b/test", "a/data"]
        self.assertFalse(self.cache.get_commit_files(sha))
        self.cache.set_commit_files(sha, files)
        self.assertEqual(self.cache.get_commit_files(sha), files)



# Copyright 2023 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import io
import os
import pathlib
import sys
import tempfile
from collections import defaultdict
from contextlib import contextmanager

from oca_port.utils import misc

from . import common


@contextmanager
def make_tmp_addon(name, manifest_name="__manifest__.py"):
    with tempfile.TemporaryDirectory() as path:
        os.system(f"mkdir -p {path}/addons/{name}")
        addons_path = f"{path}/addons"
        addon_path = f"{addons_path}/{name}"
        os.system(f"touch {addon_path}/{manifest_name}")
        yield pathlib.Path(addons_path)


@contextmanager
def capture_stdout():
    captured = io.StringIO()
    sys.stdout = captured
    yield captured
    sys.stdout = sys.__stdout__


class TestMisc(common.CommonCase):
    def test_get_manifest_path(self):
        with make_tmp_addon("foo") as addons_path:
            expected = f"{addons_path}/foo/__manifest__.py"
            assert misc.get_manifest_path(addons_path / "foo") == expected

    def test_clean_text(self):
        assert misc.clean_text("[13.0] foo 13.0") == "foo"

    def test_default_dict_from_dict(self):
        res = misc.defaultdict_from_dict({"a": 1})
        # original values preserved
        assert res["a"] == 1
        # any key is a dict by default
        assert isinstance(res["b"], defaultdict)
        assert isinstance(res["b"]["c"], defaultdict)
        assert isinstance(res["b"]["c"]["d"], defaultdict)

    def test_output_cli_mode(self):
        output = misc.Output()
        output.cli = True
        output.output = None
        with capture_stdout() as captured:
            output._print("foo")
            output._print("baz")
            output._print("bar")
            captured.seek(0)
            assert captured.read() == "foo\nbaz\nbar\n"

        assert output._render_output("json", {"a": 1}) == '{"a": 1}'

    def test_extract_ref_info_http(self):
        repo = self._git_repo(self.repo_path)
        org, repo_name, remote, branch = "Example", "test_repo", "test_ref", "16.0"
        repo.create_remote(remote, f"https://user:pwd@github.com/{org}/{repo_name}.git")
        ref = f"{remote}/{branch}"
        expected_info = {
            "ref": ref,
            "kind": "source",
            "branch": branch,
            "remote": remote,
            "repo": repo_name,
            "org": org,
            "platform": "github",
        }
        info = misc.extract_ref_info(repo, "source", ref)
        self.assertDictEqual(info, expected_info)

    def test_extract_ref_info_ssh(self):
        repo = self._git_repo(self.repo_path)
        org, repo_name, remote, branch = "Example", "test_repo", "test_ref", "16.0"
        repo.create_remote(remote, f"git@github.com:{org}/{repo_name}.git")
        ref = f"{remote}/{branch}"
        expected_info = {
            "ref": ref,
            "kind": "source",
            "branch": branch,
            "remote": remote,
            "repo": repo_name,
            "org": org,
            "platform": "github",
        }
        info = misc.extract_ref_info(repo, "source", ref)
        self.assertDictEqual(info, expected_info)



# Copyright 2024 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import hashlib
import json
import os
import pathlib
from collections import defaultdict

from . import misc


class Session:
    """Manage the user's session data, in respect to XDG conventions.

    This class is used to store the list of processed/blacklisted PRs or modules
    during a porting session.
    """

    _cache_dirname = "oca-port"
    _sessions_dirname = "sessions"

    def __init__(self, app, name):
        """Initialize user's session manager."""
        self.app = app
        self.dir_path = self._get_dir_path()
        self._sessions_dir_path = self._get_sessions_dir_path()
        self._key = hashlib.shake_256(name.encode()).hexdigest(3)
        session_file = (
            f"{self.app.addon}"
            f"-{self.app.source_version}-{self.app.target_version}"
            f"-{self._key}.json"
        )
        self._session_path = self._sessions_dir_path.joinpath(session_file)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.clear()

    @classmethod
    def _get_dir_path(cls):
        """Return the path of the session directory."""
        default_cache_dir_path = pathlib.Path.home().joinpath(".cache")
        return pathlib.Path(
            os.environ.get("XDG_CACHE_HOME", default_cache_dir_path),
            cls._cache_dirname,
        )

    def _get_sessions_dir_path(self):
        """Return the dir path storing sessions data."""
        return self.dir_path.joinpath(
            self._sessions_dirname,
            self.app.upstream_org,
        )

    def get_data(self):
        """Return the data of the session."""
        try:
            with self._session_path.open() as file_:
                return json.load(file_, object_hook=misc.defaultdict_from_dict)
        except (json.JSONDecodeError, FileNotFoundError):
            # Mainly to handle empty files (first initialization of the session)
            # but also to not crash if JSON files get corrupted.
            # Returns a "nested dict" object to not worry about checking keys
            nested_dict = lambda: defaultdict(nested_dict)  # noqa
            return nested_dict()

    def set_data(self, data):
        """Store `data` for the given `session`."""
        self._sessions_dir_path.mkdir(parents=True, exist_ok=True)
        self._save_data(data, self._session_path)

    def _save_data(self, data, path):
        try:
            with path.open(mode="w") as file_:
                json.dump(data, file_, indent=2)
        except Exception:
            pass

    def clear(self):
        """Clear the session file."""
        if self._session_path and self._session_path.exists():
            self._session_path.unlink()



# Copyright 2023 Camptocamp SA
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl)

from . import common

from oca_port.utils import session


class TestSession(common.CommonCase):
    def setUp(self):
        super().setUp()
        app = self._create_app(self.source1, self.target1, upstream_org="TEST")
        self.session_name = "test"
        self.session = session.Session(app, self.session_name)
        # Clear the session before each test
        self.session.clear()

    def tearDown(self):
        # Clear the session after each test
        self.session.clear()

    def test_session(self):
        key = "test"
        value = ["a", "b"]
        data = self.session.get_data()
        self.assertFalse(data)
        data[key] = value
        self.session.set_data(data)
        data2 = self.session.get_data()
        self.assertEqual(data2, data)



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)

import os
import hashlib
import itertools
import shutil
import tempfile
import urllib.parse
from collections import defaultdict

import click
import git
import requests

from .utils import git as g, misc
from .utils.session import Session
from .utils.misc import Output, bcolors as bc

AUTHOR_EMAILS_TO_SKIP = [
    "transbot@odoo-community.org",
    "noreply@weblate.org",
    "oca-git-bot@odoo-community.org",
    "oca+oca-travis@odoo-community.org",
    "oca-ci@odoo-community.org",
    "shopinvader-git-bot@shopinvader.com",
]

SUMMARY_TERMS_TO_SKIP = [
    "Translated using Weblate",
    "Added translation using Weblate",
]

PR_BRANCH_NAME = "oca-port-{addon}-{source_version}-to-{target_version}-{key}"

FOLDERS_TO_SKIP = [
    "setup",
    ".github",
]

FILES_TO_KEEP = [
    "requirements.txt",
    "test-requirements.txt",
    "oca_dependencies.txt",
]

BOT_FILES_TO_SKIP = [
    "README.rst",
    "static/description/index.html",
]

NEW_PR_URL = (
    "https://github.com/{from_org}/{repo_name}/compare/"
    "{to_branch}...{to_org}:{pr_branch}?expand=1&title={title}"
)

# Fake PR for commits w/o any PR (used as fallback)
FAKE_PR = g.PullRequest(*[""] * 6)


def path_to_skip(commit_path):
    """Return True if the commit path should not be ported."""
    # Allows all folders (addons!) excepted those like 'setup/' generated
    # automatically by pre-commit.
    if commit_path.isdir:
        return commit_path in FOLDERS_TO_SKIP
    # Forbid all files excepted those that developers could update
    return commit_path not in FILES_TO_KEEP


class PortAddonPullRequest(Output):
    def __init__(self, app, push_branch=True):
        """Port pull requests of an addon."""
        self.app = app
        self.push_branch = push_branch and bool(self.app.destination.remote)
        self.open_pr = bool(self.app.destination.org)
        self._results = {"process": "port_commits", "results": {}}

    def run(self):
        if not self.app.check_addon_exists_to_branch():
            if self.app.non_interactive:
                if self.app.output:
                    return False, self._render_output(self.app.output, {})
            return False, None
        self._print(
            f"{bc.BOLD}{self.app.addon}{bc.END} already exists "
            f"on {bc.BOLD}{self.app.to_branch.ref()}{bc.END}, "
            "checking PRs to port..."
        )
        branches_diff = BranchesDiff(self.app)
        branches_diff.print_diff(self.app.verbose)
        if self.app.non_interactive:
            if branches_diff.commits_diff:
                # If an output is defined we return the result in the expected format
                if self.app.output:
                    self._results["results"] = branches_diff.serialized_diff
                    return True, self._render_output(self.app.output, self._results)
                if self.app.cli:
                    # Exit with an error code if commits are eligible for (back)porting
                    # User-defined exit codes should be defined between 64 and 113.
                    # Allocate 110 for 'PortAddonPullRequest'.
                    raise SystemExit(110)
                return True, None
            if self.app.output:
                # Nothing to port -> return an empty output
                return False, self._render_output(self.app.output, {})
            return False, None
        if not self.app.dry_run:
            self._print()
            # Set a destination branch (automatically generated if not already provided)
            self.app.destination.branch = self._get_dest_branch_name(branches_diff)
            # Launch the porting session
            porting_done = self._port_pull_requests(branches_diff)
            if porting_done:
                self._commit_blacklist()
                self._push_and_open_pr()
        return True, None

    def _get_dest_branch_name(self, branches_diff):
        dest_branch_name = self.app.destination.branch
        # Define a destination branch if not set
        if branches_diff.commits_diff and not dest_branch_name:
            commits_to_port = [
                commit.hexsha
                for commit in itertools.chain.from_iterable(
                    branches_diff.commits_diff.values()
                )
            ]
            h = hashlib.shake_256("-".join(commits_to_port).encode())
            key = h.hexdigest(3)
            dest_branch_name = PR_BRANCH_NAME.format(
                addon=self.app.addon,
                source_version=self.app.source_version,
                target_version=self.app.target_version,
                key=key,
            )
        return dest_branch_name

    def _port_pull_requests(self, branches_diff):
        """Open new Pull Requests (if it doesn't exist) on the GitHub repository."""
        # Now we have a destination branch, check if there is ongoing work on it
        wip = self._print_wip_session()
        dest_branch_name = self.app.destination.branch
        # Nothing to port
        if not branches_diff.commits_diff or not dest_branch_name:
            # Nothing to port while having WIP means the porting is done
            return wip
        # Check if destination branch exists, and create it if not
        dest_branch_exists = dest_branch_name in self.app.repo.heads
        base_ref = self.app.to_branch  # e.g. 'origin/14.0'
        if dest_branch_exists:
            target_commit = self.app.repo.commit(self.app.target.ref)
            dest_commit = self.app.repo.commit(dest_branch_name)
            # If target and destination branches are on the same commit we don't care
            if target_commit != dest_commit:
                # If the local branch already exists, ask the user if he wants
                # to recreate it
                confirm = (
                    f"Branch {bc.BOLD}{dest_branch_name}{bc.END} already exists, "
                    f"recreate it from {bc.BOLD}{self.app.to_branch.ref()}{bc.END}?\n"
                    "(⚠️  you will lose ongoing work)"
                )
                if not click.confirm(confirm):
                    msg = "ℹ️  To resume the work from this branch, relaunch with:\n\n"
                    cmd = (
                        f"\t{bc.DIM}oca-port {self.app.source.ref} "
                        f"{dest_branch_name} {self.app.addon} %s{bc.END}"
                    )
                    opts = []
                    if self.app.source.branch != self.app.source_version:
                        opts.append(f"--source-version={self.app.source_version}")
                    if dest_branch_name != self.app.target_version:
                        opts.append(f"--target-version={self.app.target_version}")
                    if self.app.verbose:
                        opts.append("--verbose")
                    cmd = cmd % " ".join(opts)
                    self._print(msg + cmd)
                    return False
                # Delete the local branch
                if self.app.repo.active_branch.name == dest_branch_name:
                    # We cannot delete an active branch, checkout the underlying
                    # commit instead
                    self.app.repo.git.checkout(dest_commit)
                self.app.repo.delete_head(dest_branch_name, "-f")
                dest_branch_exists = False
                # Clear any ongoing work from the session
                session = self._init_session()
                session.clear()
        if not dest_branch_exists:
            self.app.repo.git.checkout(
                "--no-track", "-b", dest_branch_name, base_ref.ref()
            )
        # Checkout the destination branch before porting PRs
        dest_branch = g.Branch(self.app.repo, dest_branch_name)
        self.app.repo.heads[dest_branch.name].checkout()
        last_pr = (
            list(branches_diff.commits_diff.keys())[-1]
            if branches_diff.commits_diff
            else None
        )
        for pr, commits in branches_diff.commits_diff.items():
            # Check if PR has been blacklisted in user's session
            if self._is_pr_blacklisted(pr):
                if self._confirm_pr_blacklisted(pr):
                    continue
            # Port PR
            current_commit = self.app.repo.commit(dest_branch.ref())
            pr_ported = self._port_pull_request_commits(
                pr,
                commits,
                base_ref,
                dest_branch,
            )
            if pr_ported:
                # Check if commits have been ported.
                # If none has been ported, blacklist automatically the current PR.
                if self.app.repo.commit(dest_branch.ref()) == current_commit:
                    self._print("\tℹ️  Nothing has been ported, skipping")
                    self._handle_pr_blacklist(
                        pr, reason=f"(auto) Nothing to port from PR #{pr.number}"
                    )
                    msg = (
                        f"\t{bc.DIM}PR #{pr.number} has been"
                        if pr.number
                        else "Orphaned commits have been"
                    ) + f" automatically blacklisted{bc.ENDD}"
                    self._print(msg)
                    continue
                self._handle_pr_ported(pr)
                if pr == last_pr:
                    self._print("\t🎉 Last PR processed! 🎉")
        return True

    def _get_session_name(self):
        return f"{self.app.addon}-{self.app.destination.branch}"

    def _init_session(self):
        session = Session(self.app, self._get_session_name())
        data = session.get_data()
        data.setdefault("addon", self.app.addon)
        data.setdefault("repo_name", self.app.repo_name)
        data.setdefault("pull_requests", {})
        session.set_data(data)
        return session

    def _is_pr_blacklisted(self, pr):
        """Check if PR is blacklisted in current user's session."""
        session = self._init_session()
        data = session.get_data()
        return bool(data["pull_requests"]["blacklisted"][pr.ref])

    def _confirm_pr_blacklisted(self, pr):
        """Ask the user if PR should still be blacklisted."""
        self._print(
            f"- {bc.BOLD}{bc.WARNING}PR #{pr.number}{bc.END} "
            f"is blacklisted in current user's session"
        )
        if not click.confirm("\tKeep it blacklisted?"):
            # Remove the PR from the session
            session = self._init_session()
            data = session.get_data()
            if pr.ref in data["pull_requests"]["blacklisted"]:
                del data["pull_requests"]["blacklisted"][pr.ref]
            session.set_data(data)
            return False
        return True

    def _handle_pr_blacklisted(self, pr):
        """Check if PR is blacklisted in current user's session.

        Return True if workflow
        """

    def _handle_pr_blacklist(self, pr, reason=None):
        if not click.confirm("\tBlacklist this PR?"):
            return False
        if not reason:
            reason = click.prompt("\tReason", type=str)
        session = self._init_session()
        data = session.get_data()
        blacklisted = data["pull_requests"].setdefault("blacklisted", {})
        if pr.ref not in blacklisted:
            pr_data = pr.to_dict(number=True)
            pr_data["reason"] = reason
            data["pull_requests"]["blacklisted"][pr.ref] = pr_data
        session.set_data(data)
        return True

    def _handle_pr_ported(self, pr):
        session = self._init_session()
        data = session.get_data()
        ported = data["pull_requests"].setdefault("ported", {})
        if pr.ref not in ported:
            data["pull_requests"]["ported"][pr.ref] = pr.to_dict(number=True)
        session.set_data(data)

    def _commit_blacklist(self):
        session = self._init_session()
        data = session.get_data()
        blacklisted = data["pull_requests"].setdefault("blacklisted", {})
        for pr in blacklisted.values():
            if (
                self.app.storage.is_pr_blacklisted(pr["ref"])
                # TODO: Backward compat for old tracking only by number
                or self.app.storage.is_pr_blacklisted(pr["number"])
            ):
                continue
            self.app.storage.blacklist_pr(pr["ref"], reason=pr["reason"])
        if self.app.storage.dirty:
            pr_refs = ", ".join([str(pr["number"]) for pr in blacklisted.values()])
            self.app.storage.commit(
                msg=f"oca-port: blacklist PR(s) {pr_refs} for {self.app.addon}"
            )

    def _print_wip_session(self):
        session = self._init_session()
        data = session.get_data()
        wip = False
        if data["pull_requests"]:
            self._print(
                f"ℹ️  Existing session for branch "
                f"{bc.BOLD}{self.app.destination.branch}{bc.END}:"
            )
            wip = True
        # Ported PRs
        if data["pull_requests"]["ported"]:
            self._print("\t✅ Ported PRs:")
        for pr_data in data["pull_requests"]["ported"].values():
            self._print(
                f"\t- {bc.BOLD}{bc.OKBLUE}{pr_data['ref']}{bc.END} "
                f"{bc.OKBLUE}{pr_data['title']}{bc.ENDC}:"
            )
        # Blacklisted PRs
        if data["pull_requests"]["blacklisted"]:
            self._print("\t⛔ Blacklisted PRs:")
        for pr_data in data["pull_requests"]["blacklisted"].values():
            self._print(
                f"\t- {bc.BOLD}{bc.WARNING}{pr_data['ref']}{bc.END} "
                f"{bc.WARNING}{pr_data['title']}{bc.ENDC}:"
            )
        # if data["pull_requests"]:
        #     self._print()
        return wip

    def _push_and_open_pr(self):
        session = self._init_session()
        data = session.get_data()
        processed_prs = data["pull_requests"]["ported"]
        blacklisted_prs = data["pull_requests"]["blacklisted"]
        if not processed_prs and not blacklisted_prs:
            self._print("ℹ️  Nothing has been ported or blacklisted.")
            return False
        pr_data = self._prepare_pull_request_data(processed_prs, blacklisted_prs)
        # Try to push and open PR against remote repository
        is_pushed = self._push_branch_to_remote()
        if not is_pushed:
            self._print(
                f"\nℹ️  Branch {bc.BOLD}{self.app.destination.branch}{bc.END} couldn't "
                "be pushed (no remote defined)"
            )
            self._print_tips(pr_data)
            return False
        if not self.open_pr:
            self._print(
                f"\nℹ️  PR based on {bc.BOLD}{self.app.destination.branch}{bc.END} couldn't "
                "be open (no remote defined)"
            )
            self._print_tips(pr_data)
            return False
        pr_url = self._search_pull_request(pr_data["base"], pr_data["title"])
        if pr_url:
            self._print(f"Existing PR has been refreshed => {pr_url}")
        else:
            self._create_pull_request(pr_data, processed_prs)

    def _print_tips(self, pr_data):
        self._print("Here is the default PR content that would have been used:")
        self._print(f"\n{bc.BOLD}Title:{bc.END}")
        self._print(pr_data["title"])
        self._print(f"\n{bc.BOLD}Description:{bc.END}")
        self._print(pr_data["body"])

    def _port_pull_request_commits(self, pr, commits, base_ref, branch):
        """Port commits of a Pull Request in a new branch."""
        if pr.number:
            self._print(
                f"- {bc.BOLD}{bc.OKCYAN}Port PR {pr.ref}{bc.END} "
                f"{bc.OKCYAN}{pr.title}{bc.ENDC}..."
            )
            self._print(f"\t{pr.url}")
        else:
            self._print(f"- {bc.BOLD}{bc.OKCYAN}Port commits w/o PR{bc.END}...")
        # Ask the user if he wants to port the PR (or orphaned commits)
        if not click.confirm("\tPort it?" if pr.number else "\tPort them?"):
            self._handle_pr_blacklist(pr)
            return False

        # Cherry-pick commits of the source PR
        for commit in commits:
            self._print(
                f"\t\tApply {bc.OKCYAN}{commit.hexsha[:8]}{bc.ENDC} "
                f"{commit.summary}..."
            )
            # Port only relevant diffs/paths from the commit
            paths_to_port = set(commit.paths_to_port)
            for diff in commit.diffs:
                skip, message = self._skip_diff(commit, diff)
                if skip:
                    if message:
                        self._print(f"\t\t\t{message}")
                    if diff.a_path in paths_to_port:
                        paths_to_port.remove(diff.a_path)
                    if diff.b_path in paths_to_port:
                        paths_to_port.remove(diff.b_path)
                    continue
            if not paths_to_port:
                self._print("\t\t\tℹ️  Nothing to port from this commit, skipping")
                continue
            try:
                patches_dir = tempfile.mkdtemp()
                self.app.repo.git.format_patch(
                    "--keep-subject",
                    "-o",
                    patches_dir,
                    "-1",
                    commit.hexsha,
                    "--",
                    *paths_to_port,
                )
                patches = [
                    os.path.join(patches_dir, f)
                    for f in sorted(os.listdir(patches_dir))
                ]
                self.app.repo.git.am("-3", "--keep", *patches)
                shutil.rmtree(patches_dir)
            except git.exc.GitCommandError as exc:
                self._print(f"{bc.FAIL}ERROR:{bc.ENDC}\n{exc}\n")
                # High chance a conflict occurs, ask the user to resolve it
                if not click.confirm(
                    "⚠️  A conflict occurs, please resolve it and "
                    "confirm to continue the process (y) or skip this commit (N)."
                ):
                    self.app.repo.git.am("--abort")
                    continue
        return True

    @staticmethod
    def _skip_diff(commit, diff):
        """Check if a commit diff should be skipped or not.

        A skipped diff won't have its file path ported through 'git format-path'.

        Return a tuple `(bool, message)` if the diff is skipped.
        """
        if diff.deleted_file:
            if diff.a_path not in commit.paths_to_port:
                return True, ""
        if diff.b_path not in commit.paths_to_port:
            return True, ""
        if diff.renamed:
            return False, ""
        diff_path = diff.b_path.split("/", maxsplit=1)[0]
        # Skip diff updating auto-generated files (pre-commit, bot...)
        if any(file_path in diff_path for file_path in BOT_FILES_TO_SKIP):
            return (
                True,
                f"SKIP: '{diff.change_type} {diff.b_path}' diff relates "
                "to an auto-generated file, skip to avoid conflict",
            )
        # Do not accept diff on unported addons
        if (
            not misc.get_manifest_path(diff_path)
            and diff_path not in commit.addons_created
        ):
            return (
                True,
                (
                    f"{bc.WARNING}SKIP diff "
                    f"{bc.BOLD}{diff.change_type} {diff.b_path}{bc.END}: "
                    "relates to an unported addon"
                ),
            )
        if diff.change_type in ("M", "D"):
            # Do not accept update and deletion on non-existing files
            if not os.path.exists(diff.b_path):
                return (
                    True,
                    (
                        f"SKIP: '{diff.change_type} {diff.b_path}' diff relates "
                        "to a non-existing file"
                    ),
                )
        return False, ""

    def _push_branch_to_remote(self):
        """Force push the local branch to remote destination fork."""
        if not self.push_branch:
            return False
        confirm = (
            f"Push branch '{bc.BOLD}{self.app.destination.branch}{bc.END}' "
            f"to remote '{bc.BOLD}{self.app.destination.remote}{bc.END}'?"
        )
        if click.confirm(confirm):
            self.app.repo.git.push(
                self.app.destination.remote,
                self.app.destination.branch,
                "--force-with-lease",
            )
            return True
        return False

    def _prepare_pull_request_data(self, processed_prs, blacklisted_prs):
        # Adapt the content depending on the number of ported PRs
        title = body = ""
        if len(processed_prs) > 1:
            title = (
                f"[{self.app.target_version}][FW] {self.app.addon}: multiple ports "
                f"from {self.app.source_version}"
            )
            lines = [f"- #{pr['number']}" for pr in processed_prs.values()]
            body = "\n".join(
                [
                    f"Port of the following PRs from {self.app.source_version} "
                    f"to {self.app.target_version}:"
                ]
                + lines
            )
        if len(processed_prs) == 1:
            pr = list(processed_prs.values())[0]
            title = f"[{self.app.target_version}][FW] {pr['title']}"
            body = (
                f"Port of #{pr['number']} from {self.app.source_version} "
                f"to {self.app.target_version}."
            )
        # Handle blacklisted PRs
        if blacklisted_prs:
            if not title:
                title = (
                    f"[{self.app.target_version}][FW] Blacklist of some PRs "
                    f"from {self.app.source_version}"
                )
            lines2 = [
                f"- #{pr['number']}: {pr['reason']}" for pr in blacklisted_prs.values()
            ]
            body2 = "\n".join(["The following PRs have been blacklisted:"] + lines2)
            if body:
                body = "\n\n".join([body, body2])
            else:
                body = body2
        return {
            "draft": True,
            "title": title,
            "head": f"{self.app.destination.org}:{self.app.destination.branch}",
            "base": self.app.to_branch.name,
            "body": body,
        }

    def _search_pull_request(self, base_branch, title):
        params = {
            "q": (
                f"is:pr "
                f"repo:{self.app.upstream_org}/{self.app.repo_name} "
                f"base:{base_branch} "
                f"state:open {title} in:title"
            ),
        }
        response = self.app.github.request("search/issues", params=params)
        if response["items"]:
            return response["items"][0]["html_url"]

    def _create_pull_request(self, pr_data, processed_prs):
        if len(processed_prs) > 1:
            self._print(
                "PR(s) ported locally:",
                ", ".join(
                    [f"{bc.OKCYAN}#{pr.number}{bc.ENDC}" for pr in processed_prs]
                ),
            )
        if click.confirm(
            f"Create a draft PR from '{bc.BOLD}{self.app.destination.branch}{bc.END}' "
            f"to '{bc.BOLD}{self.app.to_branch.name}{bc.END}' "
            f"against {bc.BOLD}{self.app.upstream_org}/{self.app.repo_name}{bc.END}?"
        ):
            response = self.app.github.request(
                f"repos/{self.app.upstream_org}/{self.app.repo_name}/pulls",
                method="post",
                json=pr_data,
            )
            pr_url = response["html_url"]
            self._print(
                f"\t{bc.BOLD}{bc.OKCYAN}PR created =>" f"{bc.ENDC} {pr_url}{bc.END}"
            )
            return pr_url
        # Invite the user to open the PR on its own
        pr_title_encoded = urllib.parse.quote(pr_data["title"])
        new_pr_url = NEW_PR_URL.format(
            from_org=self.app.upstream_org,
            repo_name=self.app.repo_name,
            to_branch=self.app.to_branch.name,
            to_org=self.app.destination.org,
            pr_branch=self.app.destination.branch,
            title=pr_title_encoded,
        )
        self._print(
            "\nℹ️  You can still open the PR yourself there:\n" f"\t{new_pr_url}\n"
        )
        self._print_tips(pr_data)


class BranchesDiff(Output):
    """Helper to compare easily commits (and related PRs) between two branches."""

    def __init__(self, app):
        self.app = app
        self.path = self.app.addon_path
        self.from_branch_path_commits, _ = self._get_branch_commits(
            self.app.from_branch.ref(), self.path
        )
        self.from_branch_all_commits, _ = self._get_branch_commits(
            self.app.from_branch.ref()
        )
        self.to_branch_path_commits, _ = self._get_branch_commits(
            self.app.to_branch.ref(), self.path
        )
        self.to_branch_all_commits, _ = self._get_branch_commits(
            self.app.to_branch.ref()
        )
        self.commits_diff = self.get_commits_diff()
        self.serialized_diff = self._serialize_diff(self.commits_diff)
        # Once the analyze is done, we store the cache on disk
        self.app.cache.save()

    def _serialize_diff(self, commits_diff):
        data = {}
        for pr, commits in commits_diff.items():
            data[pr.number] = pr.to_dict()
            data[pr.number]["missing_commits"] = [commit.hexsha for commit in commits]
        return data

    def _get_branch_commits(self, branch, path="."):
        """Get commits from the local repository for the given `branch`.

        An optional `path` parameter can be set to limit commits to a given folder.
        This function also filters out undesirable commits (merge or translation
        commits...).

        Return two data structures:
            - a list of Commit objects `[Commit, ...]`
            - a dict of Commits objects grouped by SHA `{SHA: Commit, ...}`
        """
        commits = self.app.repo.iter_commits(branch, paths=path)
        commits_list = []
        commits_by_sha = {}
        for commit in commits:
            if self.app.cache.is_commit_ported(commit.hexsha):
                continue
            com = g.Commit(
                commit, addons_path=self.app.addons_rootdir, cache=self.app.cache
            )
            if self._skip_commit(com):
                continue
            commits_list.append(com)
            commits_by_sha[commit.hexsha] = com
        # Put ancestors at the beginning of the list to loop with
        # the expected order
        commits_list.reverse()
        return commits_list, commits_by_sha

    @staticmethod
    def _skip_commit(commit):
        """Check if a commit should be skipped or not.

        Merge or translations commits are skipped for instance, or commits
        updating only files/folders we do not want to port (pre-commit
        configuration, setuptools files...).
        """
        return (
            # Skip merge commit
            len(commit.parents) > 1
            or commit.author_email in AUTHOR_EMAILS_TO_SKIP
            or any([term in commit.summary for term in SUMMARY_TERMS_TO_SKIP])
            or all(path_to_skip(path) for path in commit.paths)
        )

    def print_diff(self, verbose=False):
        lines_to_print = []
        fake_pr = None
        i = 0
        for i, pr in enumerate(self.commits_diff, 1):
            if pr.number:
                lines_to_print.append(
                    f"{i}) {bc.BOLD}{bc.OKBLUE}{pr.ref}{bc.END} "
                    f"{bc.OKBLUE}{pr.title}{bc.ENDC}:"
                )
                lines_to_print.append(f"\tBy {pr.author}, merged at {pr.merged_at}")
            else:
                lines_to_print.append(f"{i}) {bc.BOLD}{bc.OKBLUE}w/o PR{bc.END}:")
                fake_pr = pr
            if verbose:
                pr_paths = ", ".join([f"{bc.DIM}{path}{bc.ENDD}" for path in pr.paths])
                lines_to_print.append(f"\t=> Updates: {pr_paths}")
            if pr.number:
                pr_paths_not_ported = ", ".join(
                    [f"{bc.OKBLUE}{path}{bc.ENDC}" for path in pr.paths_not_ported]
                )
                lines_to_print.append(f"\t=> Not ported: {pr_paths_not_ported}")
            lines_to_print.append(
                f"\t=> {bc.BOLD}{bc.OKBLUE}{len(self.commits_diff[pr])} "
                f"commit(s){bc.END} not (fully) ported"
            )
            if pr.number:
                lines_to_print.append(f"\t=> {pr.url}")
            if verbose or not pr.number:
                for commit in self.commits_diff[pr]:
                    lines_to_print.append(
                        f"\t\t{bc.DIM}{commit.hexsha[:8]} " f"{commit.summary}{bc.ENDD}"
                    )
        if fake_pr:
            # We have commits without PR, adapt the message
            i -= 1
            nb_commits = len(self.commits_diff[fake_pr])
            message = (
                f"{bc.BOLD}{bc.OKBLUE}{i} pull request(s){bc.END} "
                f"and {bc.BOLD}{bc.OKBLUE}{nb_commits} commit(s) w/o "
                f"PR{bc.END} related to '{bc.OKBLUE}{self.path}"
                f"{bc.ENDC}' to port from {self.app.from_branch.ref()} "
                f"to {self.app.to_branch.ref()}"
            )
        else:
            message = (
                f"{bc.BOLD}{bc.OKBLUE}{i} pull request(s){bc.END} "
                f"related to '{bc.OKBLUE}{self.path}{bc.ENDC}' to port from "
                f"{self.app.from_branch.ref()} to {self.app.to_branch.ref()}"
            )
        lines_to_print.insert(0, message)
        if self.commits_diff:
            lines_to_print.insert(1, "")
        self._print("\n".join(lines_to_print))

    def get_commits_diff(self):
        """Returns the commits which do not exist in `to_branch`, grouped by
        their related Pull Request.

        :return: a dict {PullRequest: {Commit: data, ...}, ...}
        """
        commits_by_pr = defaultdict(list)
        for commit in self.from_branch_path_commits:
            if commit in self.to_branch_all_commits:
                self.app.cache.mark_commit_as_ported(commit.hexsha)
                continue
            # Get related Pull Request if any
            pr = self._get_original_pr(commit)
            if pr:
                for pr_commit_sha in pr.commits:
                    try:
                        raw_commit = self.app.repo.commit(pr_commit_sha)
                    except ValueError:
                        # Ignore commits referenced by a PR but not present
                        # in the stable branches
                        continue
                    pr_commit = g.Commit(
                        raw_commit,
                        addons_path=self.app.addons_rootdir,
                        cache=self.app.cache,
                    )
                    if self._skip_commit(pr_commit):
                        continue
                    pr_commit_paths = {
                        path for path in pr_commit.paths if not path_to_skip(path)
                    }
                    pr.paths.update(pr_commit_paths)
                    # Check that this PR commit does not change the current
                    # addon we are interested in, in such case also check
                    # for each updated addons that the commit has already
                    # been ported.
                    # Indeed a commit could have been ported partially
                    # in the past (with git-format-patch), and we now want
                    # to port the remaining chunks.
                    if pr_commit not in self.to_branch_path_commits:
                        paths = set(pr_commit_paths)
                        # A commit could have been ported several times
                        # if it was impacting several addons and the
                        # migration has been done with git-format-patch
                        # on each addon separately
                        to_branch_all_commits = self.to_branch_all_commits[:]
                        skip_pr_commit = False
                        with g.no_strict_commit_equality():
                            while pr_commit in to_branch_all_commits:
                                index = to_branch_all_commits.index(pr_commit)
                                ported_commit = to_branch_all_commits.pop(index)
                                ported_commit_paths = {
                                    path
                                    for path in ported_commit.paths
                                    if not path_to_skip(path)
                                }
                                pr.ported_paths.update(ported_commit_paths)
                                pr_commit.ported_commits.append(ported_commit)
                                paths -= ported_commit_paths
                                if not paths:
                                    # The ported commits have already updated
                                    # the same addons than the original one,
                                    # we can skip it.
                                    skip_pr_commit = True
                        if skip_pr_commit:
                            continue
                    # We want to port commits that were still not ported
                    # for the addon we are interested in.
                    # If the commit has already been included, skip it.
                    if (
                        pr_commit in self.to_branch_path_commits
                        and pr_commit in self.to_branch_all_commits
                    ):
                        continue
                    existing_pr_commits = commits_by_pr.get(pr, [])
                    for existing_pr_commit in existing_pr_commits:
                        if (
                            existing_pr_commit == pr_commit
                            and existing_pr_commit.hexsha == pr_commit.hexsha
                        ):
                            # This PR commit has already been appended, skip
                            break
                    else:
                        commits_by_pr[pr].append(pr_commit)
            # No related PR: add the commit to the fake PR
            else:
                commits_by_pr[FAKE_PR].append(commit)
        # Sort PRs on the merge date (better to port them in the right order).
        # Do not return blacklisted PR.
        sorted_commits_by_pr = {}
        for pr in sorted(commits_by_pr, key=lambda pr: pr.merged_at or ""):
            blacklisted = self.app.storage.is_pr_blacklisted(pr.ref)
            if not blacklisted:
                # TODO: Backward compat for old tracking only by number
                blacklisted = self.app.storage.is_pr_blacklisted(pr.number)
            if blacklisted:
                msg = (
                    f"{bc.DIM}PR #{pr.number}" if pr.number else "Orphaned commits"
                ) + f" blacklisted ({blacklisted}){bc.ENDD}"
                self._print(msg)
                continue
            sorted_commits_by_pr[pr] = commits_by_pr[pr]
        return sorted_commits_by_pr

    def _get_original_pr(self, commit: g.Commit):
        """Return the original PR of a given commit."""
        # Try to get the data from the user's cache first
        data = self.app.cache.get_pr_from_commit(commit.hexsha)
        if data:
            return g.PullRequest(**data)
        # Request GitHub to get them
        if not any("github.com" in remote.url for remote in self.app.repo.remotes):
            return
        try:
            raw_data = self.app.github.get_original_pr(
                self.app.upstream_org,
                self.app.repo_name,
                self.app.from_branch.name,
                commit.hexsha,
            )
        except requests.exceptions.ConnectionError:
            self._print("⚠️  Unable to detect original PR (connection error)")
            return
        if raw_data:
            # Get all commits of the PR as they could update others addons
            # than the one the user is interested in.
            # NOTE: commits fetched from PR are already in the right order
            pr_number = raw_data["number"]
            pr_commits_data = self.app.github.request(
                f"repos/{self.app.upstream_org}/{self.app.repo_name}"
                f"/pulls/{pr_number}/commits?per_page=100"
            )
            pr_commits = [pr["sha"] for pr in pr_commits_data]
            data = {
                "number": raw_data["number"],
                "url": raw_data["html_url"],
                "author": raw_data["user"].get("login", ""),
                "title": raw_data["title"],
                "body": raw_data["body"],
                "merged_at": raw_data["merged_at"],
                "commits": pr_commits,
            }
            self.app.cache.store_commit_pr(commit.hexsha, data)
            return g.PullRequest(**data)



import json

from oca_port.utils.misc import extract_ref_info

from . import common


class TestApp(common.CommonCase):
    def test_app_init(self):
        # Simple case: call oca-port only with SOURCE, TARGET and ADDON parameters
        # Check with --dry-run option
        #   $ oca-port origin/15.0 origin/16.0 my_module --dry-run
        app = self._create_app(self.source1, self.target1, dry_run=True)
        # Check without --dry-run option
        #   $ oca-port origin/15.0 origin/16.0 my_module
        app = self._create_app(self.source1, self.target1)
        self.assertEqual(app.destination.org, app.target.org)
        self.assertEqual(app.destination.remote, app.target.remote)
        self.assertEqual(app.destination.repo, app.target.repo)
        self.assertEqual(app.destination.branch, None)  # Is automatically set later
        # Check without --dry-run option and with a different destination
        #   $ oca-port origin/15.0 origin/16.0 --destination FORK/dev my_module
        app = self._create_app(self.source1, self.target1, destination=self.destination)
        self.assertFalse(app.destination.org)
        self.assertEqual(app.destination.remote, self.fork_org)
        self.assertEqual(app.destination.branch, self.dest_branch)

        # Check with a branch that doesn't exist
        #   $ oca-port origin/15.0 14.0
        error_msg = "Ref 14.0 doesn't exist."
        with self.assertRaisesRegex(ValueError, error_msg):
            self._create_app(self.source1, "14.0")
        #   $ oca-port 14.0 origin/15.0
        with self.assertRaisesRegex(ValueError, error_msg):
            self._create_app("14.0", self.target1)

        # Check with a branch that doesn't match an Odoo version
        #   $ oca-port origin/15.0 dev
        error_msg = "Unable to identify Odoo target version from dev."
        with self.assertRaisesRegex(ValueError, error_msg):
            self._create_app(self.source1, "dev")
        #   $ oca-port dev origin/15.0
        error_msg = "Unable to identify Odoo source version from dev."
        with self.assertRaisesRegex(ValueError, error_msg):
            self._create_app("dev", self.target1)

        # Check with a local branch matching an Odoo version as target
        #   $ oca-port origin/15.0 16.0 my_module
        # NOTE: ensure the local branch exists first
        repo = self._git_repo(self.repo_path)
        repo.remotes["origin"].fetch("16.0:16.0")
        app = self._create_app(self.source1, "16.0")
        self.assertFalse(app.target.org)
        self.assertEqual(app.target.branch, "16.0")
        self.assertFalse(app.destination.org)
        self.assertFalse(app.destination.remote)
        self.assertFalse(app.destination.branch)
        self.assertFalse(app.destination.branch)

    def test_check_addon_exists(self):
        app = self._create_app(self.source1, self.target2)
        # source
        self.assertTrue(app.check_addon_exists_from_branch())
        # target
        self.assertFalse(app.check_addon_exists_to_branch())

    def test_app_nothing_to_port(self):
        app = self._create_app(self.source1, self.target1)
        try:
            app.run()
        except SystemExit as exc:
            # exit code 0 means nothing needs to be migrated/ported
            self.assertEqual(exc.args[0], 0)

    def test_app_commit_to_port(self):
        repo = self._git_repo(self.repo_path)
        source1 = extract_ref_info(repo, "source", self.source1)
        self._commit_change_on_branch(self.repo_upstream_path, source1.branch)
        app = self._create_app(self.source1, self.target1, fetch=True)
        try:
            app.run()
        except SystemExit as exc:
            # exit code 110 means pull requests or commits could be ported
            self.assertEqual(exc.args[0], 110)
        # The other way around, no commit to backport (no exception)
        # (with CLI, the returned exit code is then 0)
        app = self._create_app(self.target1, self.source1)
        res = app.run()
        self.assertFalse(res)

    def test_app_module_to_migrate(self):
        app = self._create_app(self.source2, self.target2)
        try:
            app.run()
        except SystemExit as exc:
            # exit code 100 means the module could be migrated
            self.assertEqual(exc.args[0], 100)
        # The other way around, nothing to migrate as the module doesn't exist
        # (with CLI, the returned exit code is then 1)
        app = self._create_app(self.target2, self.source2)
        error_msg = "my_module does not exist on origin/17.0"
        with self.assertRaisesRegex(ValueError, error_msg):
            app.run()

    def test_app_commit_to_port_non_interactive(self):
        repo = self._git_repo(self.repo_path)
        source1 = extract_ref_info(repo, "source", self.source1)
        self._commit_change_on_branch(self.repo_upstream_path, source1.branch)
        app = self._create_app(
            self.source1,
            self.target1,
            non_interactive=True,
            fetch=True,
        )
        result = app.run()
        self.assertTrue(result)
        self.assertIsInstance(result, bool)
        # The other way around, no commit to backport
        app = self._create_app(self.target1, self.source1, non_interactive=True)
        result = app.run()
        self.assertFalse(result)
        self.assertIsInstance(result, bool)

    def test_app_module_to_migrate_non_interactive(self):
        app = self._create_app(
            self.source2,
            self.target2,
            non_interactive=True,
        )
        result = app.run()
        self.assertTrue(result)
        self.assertIsInstance(result, bool)
        # The other way around, nothing to migrate as the module doesn't exist
        app = self._create_app(self.target2, self.source2, non_interactive=True)
        error_msg = "my_module does not exist on origin/17.0"
        with self.assertRaisesRegex(ValueError, error_msg):
            app.run()

    def test_app_wrong_output(self):
        with self.assertRaisesRegex(ValueError, "Supported outputs are"):
            self._create_app(
                self.source2,
                self.target2,
                output="wrong_format",
            )

    def test_app_commit_to_port_output_json(self):
        repo = self._git_repo(self.repo_path)
        source1 = extract_ref_info(repo, "source", self.source1)
        commit_sha = self._commit_change_on_branch(
            self.repo_upstream_path, source1.branch
        )
        app = self._create_app(
            self.source1,
            self.target1,
            output="json",
            fetch=True,
        )
        output = app.run()
        self.assertTrue(output)
        self.assertIsInstance(output, str)
        output = json.loads(output)
        self.assertEqual(output["process"], "port_commits")
        # A commit could be ported and is put in a "fake PR" without number
        self.assertEqual(len(output["results"]), 1)
        self.assertDictEqual(
            output["results"][""],
            {
                "url": "",
                "ref": "",
                "author": "",
                "title": "",
                "merged_at": "",
                "missing_commits": [commit_sha],
            },
        )
        # The other way around, no commit to backport
        app = self._create_app(self.target1, self.source1, output="json")
        output = app.run()
        self.assertTrue(output)
        self.assertIsInstance(output, str)
        output = json.loads(output)
        self.assertEqual(output, {})

    def test_app_module_to_migrate_output_json(self):
        app = self._create_app(
            self.source2,
            self.target2,
            output="json",
        )
        output = app.run()
        self.assertTrue(output)
        self.assertIsInstance(output, str)
        output = json.loads(output)
        self.assertEqual(output["process"], "migrate")
        self.assertEqual(output["results"], {})
        # The other way around, nothing to migrate as the module doesn't exist
        app = self._create_app(self.target2, self.source2, output="json")
        error_msg = "my_module does not exist on origin/17.0"
        with self.assertRaisesRegex(ValueError, error_msg):
            app.run()



# Copyright 2022 Camptocamp SA
# License LGPL-3.0 or later (http://www.gnu.org/licenses/lgpl)
import os
import pathlib
from dataclasses import dataclass
import re

import git

from . import utils
from .exceptions import RemoteBranchValueError
from .migrate_addon import MigrateAddon
from .port_addon_pr import PortAddonPullRequest
from .utils.git import Branch
from .utils.github import GitHub
from .utils.misc import Output, bcolors as bc, extract_ref_info


@dataclass
class App(Output):
    """'oca-port' application centralizing settings and operations.

    Parameters:

        source:
            string representation of the source branch, e.g. 'origin/15.0'
        target:
            string representation of the target branch, e.g. 'origin/16.0'
        addon_path:
            the path of the addon to process
        destination:
            string representation of the destination branch,
            e.g. 'camptocamp/16.0-addon-dev'
        source_version:
            Source Odoo version. To set if it cannot be detected from 'source'.
        target_version:
            Target Odoo version. To set if it cannot be detected from 'target'.
        repo_path:
            local path to the Git repository
        repo_name:
            name of the repository on the upstream organization (e.g. 'server-tools')
        verbose:
            returns more details to the user
        non_interactive:
            flag to not wait for user input and to return a error code to the shell.
            Returns 100 if an addon could be migrated, 110 if pull requests/commits
            could be ported, 0 if the history of the addon is the same on both branches.
        output:
            returns a parsable output. This implies the 'non-interactive' mode
            defined above but without returning any special exit code.
            Possible values: 'json'
        fetch:
            always fetch source and target branches from upstream
        no_cache:
            flag to disable the user's cache
        clear_cache:
            flag to remove the user's cache once the process is done
        github_token:
            Token to use when requesting GitHub API (highly recommended
            to not trigger the "API rate limit exceeded" error).
    """

    source: str
    target: str
    addon_path: str
    destination: str = None
    source_version: str = None
    target_version: str = None
    repo_path: str = ""
    repo_name: str = None
    upstream_org: str = "OCA"
    verbose: bool = False
    non_interactive: bool = False
    dry_run: bool = False
    output: str = None
    fetch: bool = False
    no_cache: bool = False
    clear_cache: bool = False
    github_token: str = None
    cli: bool = False  # Not documented, should not be used outside of the CLI

    _available_outputs = ("json",)

    def __post_init__(self):
        self._prepare_parameters()
        # Force non-interactive mode:
        #   - if we are not in CLI mode
        if not self.cli:
            self.non_interactive = True
        #   - if an output has been defined
        if self.output:
            if self.output.lower() not in self._available_outputs:
                outputs = ", ".join(self._available_outputs)
                raise ValueError(f"Supported outputs are: {outputs}")
            self.non_interactive = True
        # Fetch branches if they can't be resolved locally
        # NOTE: required for the storage below to retrieve data
        remote_branches = self.repo.git.branch("-r").split()
        if (
            self.fetch
            or (
                self.from_branch.remote
                and self.from_branch.ref() not in remote_branches
            )
            or (self.to_branch.remote and self.to_branch.ref() not in remote_branches)
        ):
            self.fetch_branches()
        # Check if source & target branches exist
        self._check_branch_exists(self.source.ref, raise_exc=True)
        self._check_branch_exists(self.target.ref, raise_exc=True)
        # GitHub API helper
        self.github = GitHub(self.github_token or os.environ.get("GITHUB_TOKEN"))
        # Initialize storage & cache
        self.storage = utils.storage.InputStorage(self.to_branch, self.addon)
        self.cache = utils.cache.UserCacheFactory(self).build()

    def _handle_odoo_versions(self):
        odoo_version_pattern = r"^[0-9]+\.[0-9]$"
        source_version = re.search(odoo_version_pattern, self.source.branch)
        target_version = re.search(odoo_version_pattern, self.target.branch)
        source_param = "--source-version" if self.cli else "source_version"
        target_param = "--target-version" if self.cli else "target_version"
        # Check Odoo versions from branches
        if not source_version and not self.source_version:
            raise ValueError(
                f"Unable to identify Odoo source version from {self.source.branch}.\n"
                f"Use {source_param} parameter to identify Odoo source version."
            )
        if not target_version and not self.target_version:
            raise ValueError(
                f"Unable to identify Odoo target version from {self.target.branch}.\n"
                f"Use {target_param} parameter to identify target Odoo version."
            )
        # Check source_version and target_version parameters
        if self.source_version and not re.search(
            odoo_version_pattern, self.source_version
        ):
            raise ValueError(f"Unable to identify Odoo version from {source_param}.")
        if self.target_version and not re.search(
            odoo_version_pattern, self.target_version
        ):
            raise ValueError(f"Unable to identify Odoo version from {target_param}.")
        self.source_version = self.source_version or source_version.string
        self.target_version = self.target_version or target_version.string

    def _prepare_parameters(self):
        # Handle Git repository
        self.repo = git.Repo(self.repo_path)
        if self.repo.is_dirty(untracked_files=True):
            raise ValueError("changes not committed detected in this repository.")

        # Module name
        self.addon_path = pathlib.Path(self.addon_path)
        self.addons_rootdir = self.addon_path.parent
        self.addon = self.addon_path.name

        # Convert them to full remote info if needed
        for key in ("source", "target", "destination"):
            value = getattr(self, key)
            if value and isinstance(value, str):
                setattr(self, key, extract_ref_info(self.repo, key, value))
        # Check Odoo versions from source and target branches and parameters
        self._handle_odoo_versions()

        # Always provide a destination:
        if not self.destination:
            self.destination = extract_ref_info(self.repo, "destination", "")
            # Unset org that could have been taken from 'origin'.
            self.destination.org = None
            # If target.org is different than upstream_org, generate the
            # destination from target so we get the remote+org for free (if any)
            if self.target.org != self.upstream_org:
                self.destination = extract_ref_info(
                    self.repo, "destination", self.target.ref
                )
                # If destination is not a local one, or not an Odoo version,
                # it will be generated by the specific tool
                if (
                    self.destination.remote
                    or self.destination.branch == self.target_version
                ):
                    self.destination.branch = None

        # Handle with repo_path and repo_name
        self.repo_path = pathlib.Path(self.repo_path)
        self.repo_name = (
            self.repo_name
            or self.source.repo
            or self.target.repo
            or self.repo_path.absolute().name
        )
        if not self.repo_path:
            raise ValueError("'repo_path' has to be set.")

        # Transform branch strings to Branch objects
        self.from_branch = self._prepare_branch(self.source)
        self.to_branch = self._prepare_branch(self.target)
        self.dest_branch = self.to_branch
        if self.destination.branch:
            self.dest_branch = self._prepare_branch(self.destination)

        # Print a warning if the repository in the destination remote URL differs
        # from the repository of the target as we probably want to push under the
        # same repository name.
        fishy_parameters = False
        if not self.dry_run and self.destination.remote and self.target.repo:
            dest_remote_url = self.repo.remotes[self.destination.remote].url
            if self.target.repo not in dest_remote_url:
                fishy_parameters = True
                self._print(
                    "⚠️  Destination repository does not match the target. "
                    "Destination remote is perhaps not the right one."
                )
        # Print a summary of source/target/destination
        if self.verbose or fishy_parameters:
            source_remote_url = (
                self.repo.remotes[self.source.remote].url if self.source.remote else ""
            )
            msg = f"{bc.BOLD}Source{bc.END}: {self.source.ref}"
            if source_remote_url:
                msg += f", remote {bc.BOLD}{self.source.remote} {source_remote_url}{bc.END}"
            self._print(msg)
            target_remote_url = (
                self.repo.remotes[self.target.remote].url if self.target.remote else ""
            )
            msg = f"{bc.BOLD}Target{bc.END}: {self.target.ref}"
            if target_remote_url:
                msg += f", remote {bc.BOLD}{self.target.remote} {target_remote_url}{bc.END}"
            self._print(msg)
            if not self.dry_run:
                dest_remote_url = (
                    self.repo.remotes[self.destination.remote].url
                    if self.destination.remote
                    else ""
                )
                msg = f"{bc.BOLD}Destination{bc.END}: {self.destination.ref}"
                if dest_remote_url:
                    msg += (
                        f", remote {bc.BOLD}{self.destination.remote} "
                        f"{dest_remote_url}{bc.END}"
                    )
                self._print(msg)

    def _prepare_branch(self, info):
        try:
            return Branch(self.repo, info.branch, default_remote=info.remote)
        except ValueError as exc:
            remote = exc.args[1]
            if remote not in self.repo.remotes:
                raise RemoteBranchValueError(info) from exc

    def fetch_branches(self):
        for branch in (self.from_branch, self.to_branch):
            if not branch.remote:
                continue
            remote_url = branch.repo.remotes[branch.remote].url
            if self.verbose:
                self._print(f"Fetch {bc.BOLD}{branch.ref()}{bc.END} from {remote_url}")
            branch.repo.remotes[branch.remote].fetch(branch.name)

    def _check_branch_exists(self, branch, raise_exc=False):
        for ref in self.repo.refs:
            if branch == ref.name:
                return True
        if raise_exc:
            raise ValueError(f"Ref {branch} doesn't exist.")
        return False

    def _check_addon_exists(self, branch, raise_exc=False):
        repo = self.repo
        addons_tree = repo.commit(branch.ref()).tree
        if self.addons_rootdir and self.addons_rootdir.name:
            addons_tree /= str(self.addons_rootdir)
        branch_addons = [t.path for t in addons_tree.trees]
        if str(self.addon_path) not in branch_addons:
            if not raise_exc:
                return False
            error = f"{self.addon_path} does not exist on {branch.ref()}"
            if self.cli:
                error = (
                    f"{bc.FAIL}{self.addon_path}{bc.ENDC} "
                    "does not exist on {branch.ref()}"
                )
            raise ValueError(error)
        return True

    def check_addon_exists_from_branch(self, raise_exc=False):
        """Check that `addon` exists on the source branch`."""
        return self._check_addon_exists(self.from_branch, raise_exc=raise_exc)

    def check_addon_exists_to_branch(self, raise_exc=False):
        """Check that `addon` exists on the target branch`."""
        return self._check_addon_exists(self.to_branch, raise_exc=raise_exc)

    def run(self):
        """Run 'oca-port' to migrate an addon or to port its pull requests."""
        self.check_addon_exists_from_branch(raise_exc=True)
        # Check if some PRs could be ported
        res, output = self.run_port()
        if not res:
            # If not, migrate the addon
            res, output = self.run_migrate()
        if self.cli and self.output:
            if not output:
                output = self._render_output(self.output, {})
            print(output)
        if self.clear_cache:
            self.cache.clear()
        if self.output:
            return output
        return res

    def run_port(self):
        """Port pull requests of an addon (if any)."""
        # Check if the addon (folder) exists on the target branch
        #   - if it already exists, check if some PRs could be ported
        return PortAddonPullRequest(self).run()

    def run_migrate(self):
        """Migrate an addon."""
        return MigrateAddon(self).run()
